from aiohttp_test_utils import init_tests
from jdatetime import datetime as jdatetime

# noinspection PyProtectedMember
from tsetmc import _MarketState

init_tests()


def assert_market_state(market_state: _MarketState):
    assert type(market_state.pop('datetime')) is jdatetime
    for k in ('tse_status', 'fb_status', 'derivatives_status'):
        assert type(market_state.pop(k)) is str
    for k in ('fb_tno', 'derivatives_tno'):
        assert type(market_state.pop(k)) is int
    # tse_value is optional
    assert type(market_state.pop('tse_value', 0.0)) is float
    assert market_state.keys() == {
        'tse_index', 'tse_index_change', 'tse_tvol', 'tse_tval', 'tse_tno',
        'fb_tvol', 'fb_tval', 'derivatives_tvol', 'derivatives_tval',
        'tse_index_change_percent'}
    assert all(type(v) is float for v in market_state.values())


def assert_dict_type(d: dict, td: callable):
    assert td.__optional_keys__ == (d.keys() - td.__required_keys__)
    for k, t in td.__annotations__.items():
        v = d[k]
        if isinstance(v, dict):
            assert_dict_type(v, t)
            continue
        assert isinstance(v, t), f'{td=} {k=} {v=} {t=}'



from asyncio import sleep as _sleep
from collections.abc import Callable as _Callable
from io import BytesIO as _BytesIO, StringIO as _StringIO
from logging import error as _erorr
from typing import Any as _Any

from numpy import nan as _nan
from pandas import read_html as _read_html, to_numeric as _to_numeric

from tsetmc import (
    _csv2df,
    _DataFrame,
    _get_data,
    _get_par_tree,
    _jstrptime,
    _MarketState,
    _parse_market_state,
    _parse_ombud_messages,
    _TypedDict,
)

_BEST_LIMITS_NAMES = ('ins_code', 'number', 'zo', 'zd', 'pd', 'po', 'qd', 'qo')
_PRICE_DTYPES = {
    'ins_code': 'string',
    'isin': 'string',
    'l18': 'string',
    'l30': 'string',
    'heven': 'int32',
    'pf': 'int64',
    'pc': 'int64',
    'pl': 'int64',
    'tno': 'int64',
    'tvol': 'int64',
    'tval': 'int64',
    'pmin': 'int64',
    'pmax': 'int64',
    'py': 'int64',
    'eps': 'float64',
    'bvol': 'int64',
    'visitcount': 'int64',
    # 0-7 http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
    'flow': 'int16',
    # 1-98, see tsetmc.general.cs_codes()
    'cs': 'int16',
    'tmax': 'float64',
    'tmin': 'float64',
    'z': 'int64',
    # 67-701 http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
    'yval': 'int16',
    'predtran': 'float64',
    'buyop': 'Int64',
}
_PRICE_COLUMNS = _PRICE_DTYPES.keys()
_PRICE_UPDATE_COLUMNS = ('ins_code', *(*_PRICE_COLUMNS,)[4:13])


class _MarketWatchInit(_TypedDict, total=False):
    prices: _DataFrame
    best_limits: _DataFrame
    market_state: _MarketState
    refid: int


async def market_watch_init(
    *, market_state=True, prices=True, best_limits=True, join=True
) -> _MarketWatchInit:
    """Return the market status which are the info used in creating filters.

    If `join` is True, `best_limits` will be joined to `prices`, otherwise
        it will be returned as a separate dataframes.

    For more information about filters see:
        http://tsetmc.com/Loader.aspx?ParTree=15131F
    For the meaning of column names in the returned DataFrame refer to:
        http://tsetmc.com/Site.aspx?ParTree=151713
        For persian translation of `flow` and `yval` codes use
        `tsetmc.docs.instrument`.
        `heven` is the time of the last transaction in HHMMSS format.
    """
    # todo: use content?
    text = await _get_data('MarketWatchInit.aspx?h=0&r=0', fa=True)
    _, market_state_str, states, price_rows, refid = text.split('@')
    result = {'refid': int(refid)}
    if prices:
        result['prices'] = price_df = _csv2df(
            _StringIO(states),
            names=_PRICE_COLUMNS,
            index_col='ins_code',
            dtype=_PRICE_DTYPES,
        )
    if best_limits:
        result['best_limits'] = best_limits_df = _csv2df(
            _StringIO(price_rows),
            names=_BEST_LIMITS_NAMES,
            dtype={'ins_code': 'string'},
            index_col=('ins_code', 'number'),
        )
    if join and prices and best_limits:
        # merge multiple rows sharing the same `row` number into one row.
        # a fascinating solution from https://stackoverflow.com/a/53563551/2705757
        # noinspection PyUnboundLocalVariable
        best_limits_df = best_limits_df.unstack(fill_value=0).sort_index(
            axis=1, level=1
        )
        best_limits_df.columns = [f'{c}{i}' for c, i in best_limits_df.columns]
        # noinspection PyUnboundLocalVariable
        joined = best_limits_df.join(price_df)
        # joined_df.index = to_numeric(joined_df.index, downcast='unsigned')
        result['prices'] = joined
    if market_state:
        result['market_state'] = _parse_market_state(market_state_str)
    return result


class _MarketWatchPlus(_TypedDict, total=False):
    new_prices: _DataFrame
    price_updates: _DataFrame
    best_limits: _DataFrame
    messages: list[str]
    market_state: _MarketState
    refid: int


async def market_watch_plus(
    heven: int,
    refid: int,
    *,
    messages=True,
    market_state=True,
    new_prices=True,
    price_updates=True,
    best_limits=True,
) -> _MarketWatchPlus:
    # See dev/tsetmc_source_files/market_watch.html
    # for how the response is parsed in the browser.
    text = await _get_data(
        f'MarketWatchPlus.aspx?h={5 * (heven // 5)}&r={25 * (refid // 25)}',
        fa=True,
    )
    try:
        (
            handle_msg,
            update_fast_view,
            inst_price,
            best_limit,
            refid,
        ) = text.split('@')
    except ValueError:
        _erorr(f'{text = }')
        raise
    result = {}
    if messages:
        # whenever a new id appears, users should try to fetch new messages
        # using relevant functions
        # todo: implement functions to fetch messages using message ids
        # NewMsgNotification, NewInsStateNotification, NewCodalNotification
        result['messages'] = handle_msg.split(',')
    if market_state:
        if update_fast_view != '':
            result['market_state'] = _parse_market_state(update_fast_view)
    if new_prices or price_updates:
        inst_prices = [ip.split(',') for ip in inst_price.split(';')]
        if new_prices:
            lst = [ip for ip in inst_prices if len(ip) != 10]
            try:
                df = _DataFrame(lst, columns=_PRICE_COLUMNS, copy=False)
            except ValueError as e:
                _erorr(f'{lst[0] = }')
                raise e
            df['eps'].replace('', _nan, inplace=True)
            df['predtran'].replace('', _nan, inplace=True)
            df['buyop'].replace('', _nan, inplace=True)
            df = df.astype(_PRICE_DTYPES, False)
            df.set_index('ins_code', inplace=True)
            result['new_prices'] = df
        if price_updates:
            lst = [ip for ip in inst_prices if len(ip) == 10]
            df = _DataFrame(lst, columns=_PRICE_UPDATE_COLUMNS, copy=False)
            df.ins_code = df.ins_code.astype('string')
            df.set_index('ins_code', inplace=True)
            df = df.astype('int64', False)
            result['price_updates'] = df
    if best_limits:
        result['best_limits'] = _csv2df(
            _StringIO(best_limit),
            index_col='ins_code',
            names=_BEST_LIMITS_NAMES,
            dtype={'ins_code': 'string'},
        )
    result['refid'] = int(refid)
    return result


def _split_id_rows(content: bytes, id_row_len: int) -> list:
    data = content.split(b';')
    for i, datum in enumerate(data):
        items = datum.split(b',')
        if len(items) == id_row_len:
            id_ = items[0]
        else:
            # noinspection PyUnboundLocalVariable
            items.insert(0, id_)
        # noinspection PyTypeChecker
        data[i] = items
    return data


async def closing_price_all() -> _DataFrame:
    """Return price history dataframe.

    For the meaning of column names refer to
        http://cdn.tsetmc.com/Site.aspx?ParTree=151713
    This method returns the same set of information as `[ih]` variable in
        tsetmc filters. See:
            http://tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3197
    """
    content = await _get_data('ClosingPriceAll.aspx')
    data = _split_id_rows(content, id_row_len=11)
    columns = [
        'ins_code',
        'n',
        'pc',
        'pl',
        'tno',
        'tvol',
        'tval',
        'pmin',
        'pmax',
        'py',
        'pf',
    ]
    df = _DataFrame(
        data,
        columns=columns,
        copy=False,
    )
    df[columns[1:]] = df[columns[1:]].apply(_to_numeric)
    df.ins_code = df.ins_code.astype('string')
    df.set_index(['ins_code', 'n'], inplace=True)
    return df


async def client_type_all() -> _DataFrame:
    """Return client types (natural/legal stats) as a DataFrame.

    In column names `n_` prefix stands for natural and `l_` for legal.
    """
    content = await _get_data('ClientTypeAll.aspx')
    df = _csv2df(
        _BytesIO(content),
        names=(
            'ins_code',
            'n_buy_count',
            'l_buy_count',
            'n_buy_volume',
            'l_buy_volume',
            'n_sell_count',
            'l_sell_count',
            'n_sell_volume',
            'l_sell_volume',
        ),
        index_col='ins_code',
        dtype={'ins_code': 'string'},
    )
    return df


async def key_stats() -> _DataFrame:
    """Return key statistics as a DataFrame.

    For the meaning of column names refer to
        http://www.tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3199 or
        http://cdn.tsetmc.com/Site.aspx?ParTree=151713
    """
    content = await _get_data('InstValue.aspx?t=a')
    data = _split_id_rows(content, id_row_len=3)
    df = _DataFrame(data, columns=('ins_code', 'n', 'value'), copy=False)
    df.set_index(df.pop('ins_code').astype('string'), inplace=True)
    df = df.apply(_to_numeric)
    df = df.pivot(columns='n', values='value')
    df.columns = [f'is{c}' for c in df.columns]
    return df


async def ombud_messages(
    *,
    top: int | str = None,
    flow: int | str = None,
    containing: str = None,
    sh_date: str = None,
) -> _DataFrame:
    """Return ombud messages as a dataframe.

    :param top: How many messages to return.
    :param flow: Only return messages in the given market flow.
    :param containing: Only return messages containing this term.
    :param sh_date: Solar Hijri date string in 'YYYY-mm-dd' format.

    Note: the server ignores the `top` when `deven_persian` is given.
    """
    path = '151313'
    if flow is not None:
        path += f'&Flow={flow}'
    if top is not None:
        path += f'&top={top}'
    if sh_date is not None:
        path += f'&DevenPersian={sh_date}'
    if containing is not None:
        path += f'&Lval18AFC={containing}'
    return _parse_ombud_messages(await _get_par_tree(path))


async def status_changes(top: int | str) -> _DataFrame:
    text = await _get_par_tree(f'15131L&top={top}')
    df = _read_html(text)[0]
    df['date'] = (df['تاریخ'] + ' ' + df['زمان']).apply(
        _jstrptime, format='%Y/%m/%d %H:%M:%S'
    )
    df.drop(columns=['تاریخ', 'زمان'], inplace=True)
    return df


class MarketWatch:
    __slots__ = (
        'interval',
        'init_callback',
        'init_kwargs',
        'plus_callback',
        'plus_kwargs',
        'refid',
        'heven',
    )

    def __init__(
        self,
        *,
        init_kwargs: dict = None,
        plus_kwargs: dict = None,
        init_callback: _Callable[[_MarketWatchInit], _Any],
        plus_callback: _Callable[[_MarketWatchPlus], _Any],
        interval=1,
    ):
        """Create an object that helps with watching the market watch.

        :param init_kwargs: kwargs to be passed to market_watch_init
        :param plus_kwargs: kwargs to be passed to market_watch_plus
        :param init_callback: function to be called with the result of
            market_watch_init. This function should return True, otherwise the
            watch will be stopped.
        :param plus_callback: function to be called with the result of
            market_watch_plus. This function should return True, otherwise the
            watch will be stopped.
        :param interval: The sleep interval.

        To stop the watch, set `plus_callback` to `lambda a: False`.
        """
        self.interval = interval
        self.init_kwargs: dict = {} if init_kwargs is None else init_kwargs
        self.init_callback = init_callback
        self.plus_kwargs: dict = {} if plus_kwargs is None else plus_callback
        self.plus_callback = plus_callback

    async def start(self):
        mwi = await market_watch_init(**self.init_kwargs)
        if not self.init_callback(mwi):
            return

        heven = mwi['prices'].heven.max()
        refid = mwi['refid']

        while True:
            await _sleep(self.interval)
            try:
                mwp = await market_watch_plus(
                    refid=refid, heven=heven, **self.plus_kwargs
                )
            except Exception as e:
                _erorr(f'Exception awaiting market_watch_plus: %s', e)
                continue  # _sleep and retry
            if not self.plus_callback(mwp):
                return
            refid = mwp['refid']
            heven = max(
                mwp['price_updates'].heven.max(),
                mwp['new_prices'].heven.max(),
            )



from typing import TypedDict as _TypedDict

from bs4 import BeautifulSoup as _BeautifulSoup
from pandas import (
    NA as _NA,
    concat as _concat,
    json_normalize as _json_normalize,
    read_html as _read_html,
)

from tsetmc import _api, _DataFrame, _get_par_tree, _numerize, _partial

_make_soup = _partial(_BeautifulSoup, features='lxml')


async def boards() -> dict[int, str]:
    """See http://en.tsetmc.com/Loader.aspx?ParTree=121C1913."""
    content = await _get_par_tree('111C1913')
    iloc = _read_html(content, header=0)[0].iloc
    return dict(zip(iloc[:, 0], iloc[:, 1]))


async def cs_codes() -> dict[str, str]:
    """http://www.tsetmc.com/Loader.aspx?ParTree=111C1213"""
    content = await _get_par_tree('111C1213')
    iloc = _read_html(content, header=0)[0].iloc
    return dict(zip(iloc[:, 0], iloc[:, 1]))


async def industrial_groups_overview() -> _DataFrame:
    """Return a dataframe of industrial groups.

     The result contains info about each group's price change.
     See: http://www.tsetmc.com/Loader.aspx?ParTree=111C1214
     """
    content = await _get_par_tree('111C1214')
    df = _read_html(content)[0]
    show = df[1]
    df.drop(columns=1, inplace=True)
    percents = show.str.extract(
        r"showBar\('[^\[]*',(\d+),(\d+),(\d+),(\d+)\)"
    ).astype('int64')
    df = _concat((df, percents), copy=False, axis=1)
    df.columns = ('group', ':-2', '-2:0', '0:2', '2:')
    return df


async def market_map_data(*, market=0, size=9999, sector=0, typeSelected=1, heven=0) -> _DataFrame:
    j = await _api(
        f'ClosingPrice/GetMarketMap'
        f'?market={market}&size={size}&sector={sector}'
        f'&typeSelected={typeSelected}&hEven={heven}',
        fa=True,
    )
    df = _DataFrame(j)
    df['insCode'] = df.insCode.astype('int64')
    return df


async def major_holders_activity() -> _DataFrame:
    text = await _get_par_tree('15131I')
    soup = _make_soup(text)
    trs = soup.select('tr')

    rows = []
    append_row = rows.append
    for tr in trs[1:]:
        tds = tr.select('td')
        td0 = tds[0]

        inst_div = td0.select_one('div')
        if inst_div:
            href = inst_div.select_one('a')['href']
            ins_code = int(href[href.rfind('=') + 1:])
            l30 = inst_div.text

        holder = td0.select_one('li').text
        # noinspection PyUnboundLocalVariable
        append_row([ins_code, l30, holder, *_parse_tds(tds)])
    return _DataFrame(rows, copy=False, columns=(
        'ins_code', 'l30', 'holder', *(
            # the first header is 'شرکت - سهامدار'
            th.text for th in trs[0].select('th')[1:]
        )
    ))


async def top_industry_groups() -> _DataFrame:
    """http://www.tsetmc.com/Loader.aspx?Partree=15131O"""
    text = await _get_par_tree('15131O')
    df = _read_html(text)[0]
    df.columns = ['group', 'mv', 'tno', 'tvol', 'tval']
    _numerize(df, ('mv', 'tvol', 'tval'), float, comma=True)
    return df


def _parse_tds(tds):
    for td in tds[1:]:
        text = td.text
        try:
            yield float(text.replace(',', ''))
        except ValueError:
            if td == '\xa0':
                yield _NA


class _MarketOverview(_TypedDict):
    lastDataDEven: int
    lastDataHEven: int
    indexLastValue: float
    indexChange: float
    indexEqualWeightedLastValue: float
    indexEqualWeightedChange: float
    marketActivityDEven: int
    marketActivityHEven: int
    marketActivityZTotTran: int
    marketActivityQTotCap: float
    marketActivityQTotTran: float
    marketState: str
    marketValue: float
    marketValueBase: float
    marketStateTitle: str


async def market_overview(n=1) -> _MarketOverview:
    """Return GetMarketOverview result.

    :param n:
        1: bourse
        2: fara-bourse
    """
    j = await _api(f'MarketData/GetMarketOverview/{n}')
    return j['marketOverview']


async def related_companies(cs: str) -> dict[str, _DataFrame]:
    j = await _api(f'ClosingPrice/GetRelatedCompany/{cs}')
    j['relatedCompany'] = _json_normalize(j['relatedCompany'])
    j['relatedCompanyThirtyDayHistory'] = _DataFrame(j['relatedCompanyThirtyDayHistory'], copy=False)
    return j


from aiohttp_test_utils import file
from numpy import dtype
from pytest import warns

from tests import assert_dict_type
from tsetmc.instruments import Instrument, _ClientTypeOnDate, _ClosingPrice

FARAZ_ON_DATE = Instrument(13666407494621646).on_date(20220222)


@file('faraz_GetClosingPriceHistory_20220222.json')
async def test_intraday_closing_price():
    df = await FARAZ_ON_DATE.closing_price_history()
    assert [*df.dtypes.items()] == [
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
    ]


@file('faraz_GetStaticThreshold_20220222.json')
async def test_static_thresholds():
    df = await FARAZ_ON_DATE.static_thresholds()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('psGelStaMax', dtype('float64')),
        ('psGelStaMin', dtype('float64')),
    ]


@file('faraz_BestLimits_20220222.json')
async def test_intraday_best_limits():
    df = await FARAZ_ON_DATE.best_limits()
    assert [*df.dtypes.items()] == [
        ('idn', dtype('int64')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('refID', dtype('int64')),
        ('number', dtype('int64')),
        ('qTitMeDem', dtype('int64')),
        ('zOrdMeDem', dtype('int64')),
        ('pMeDem', dtype('float64')),
        ('pMeOf', dtype('float64')),
        ('zOrdMeOf', dtype('int64')),
        ('qTitMeOf', dtype('int64')),
        ('insCode', dtype('O')),
    ]


@file('faraz_GetTradeHistory_20220222.json')
async def test_intraday_trades():
    df = await FARAZ_ON_DATE.trades()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('nTran', dtype('int64')),
        ('hEven', dtype('int64')),
        ('qTitTran', dtype('int64')),
        ('pTran', dtype('float64')),
        ('qTitNgJ', dtype('int64')),
        ('iSensVarP', dtype('O')),
        ('pPhSeaCotJ', dtype('float64')),
        ('pPbSeaCotJ', dtype('float64')),
        ('iAnuTran', dtype('int64')),
        ('xqVarPJDrPRf', dtype('float64')),
        ('canceled', dtype('int64')),
    ]


@file('faraz_GetInstrumentHistory_20220222.json')
async def test_historic_data():
    d = await FARAZ_ON_DATE.data()
    assert d.keys() == {
        'insCode',
        'lVal30',
        'lVal18AFC',
        'flow',
        'cIsin',
        'zTitad',
        'baseVol',
        'instrumentID',
        'cgrValCot',
        'cComVal',
        'lastDate',
        'sourceID',
        'flowTitle',
        'cgrValCotTitle',
    }


@file('faraz_GetClosingPriceDaily_20220222.json')
async def test_historic_data():
    d = await FARAZ_ON_DATE.closing_price()
    assert_dict_type(d, _ClosingPrice)


@file('faraz_GetInstrumentState_20220222.json')
async def test_intraday_states():
    df = await FARAZ_ON_DATE.states()
    assert [*df.dtypes.items()] == [
        ('idn', dtype('int64')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('insCode', dtype('O')),
        ('lVal18AFC', dtype('O')),
        ('lVal30', dtype('O')),
        ('cEtaval', dtype('O')),
        ('realHeven', dtype('int64')),
        ('underSupervision', dtype('int64')),
        ('cEtavalTitle', dtype('O')),
    ]


@file('faraz_Shareholder_20220222.json')
async def test_intraday_holders():
    df = await FARAZ_ON_DATE.holders()
    assert [*df.dtypes.items()] == [
        ('shareHolderID', dtype('int64')),
        ('shareHolderName', dtype('O')),
        ('cIsin', dtype('O')),
        ('dEven', dtype('int64')),
        ('numberOfShares', dtype('float64')),
        ('perOfShares', dtype('float64')),
        ('change', dtype('int64')),
        ('changeAmount', dtype('float64')),
        ('shareHolderShareID', dtype('int64')),
    ]


@file('ondate_client_type.json')
async def test_client_type():
    d = await FARAZ_ON_DATE.client_type()
    assert_dict_type(d, _ClientTypeOnDate)
    with warns():
        d2 = await FARAZ_ON_DATE.client_types()
    assert d2 == d



__version__ = '0.51.0'

from functools import partial as _partial
from json import JSONDecodeError, loads
from logging import error
from re import compile as _rc, findall as _findall
from typing import TypedDict as _TypedDict
from warnings import warn as _warn

from aiohttp import (
    ClientSession as _ClientSession,
    ClientTimeout as _ClientTimeout,
)
from jdatetime import datetime as _jdatetime
from pandas import DataFrame as _DataFrame, read_csv as _read_csv

_csv2df = _partial(_read_csv, low_memory=False, engine='c', lineterminator=';')
_F = r'(-?\d+(?:\.\d+)?)'  # float pattern


_INDEX_CHANGE_MATCH = _rc(rf"<div[^>]*>(\()?{_F}\)?</div>(?: {_F}%)?").match
_INDEX_TIMESTAMP_MATCH = _rc(r'(\d\d)/(\d+)/(\d+) (\d\d):(\d\d):(\d\d)').match


_jstrptime = _jdatetime.strptime


class _MarketState(_TypedDict, total=False):
    datetime: _jdatetime
    tset_status: str
    tset_index: float
    tset_index_change: float
    tse_tvol: float
    tse_tval: float
    tse_tno: float
    fb_status: str
    fb_tvol: float
    fb_tval: float
    fb_tno: int
    derivatives_status: int


class _Eps(_TypedDict):
    epsValue: None
    estimatedEPS: str | None
    sectorPE: float
    psr: float


class _Sector(_TypedDict):
    dEven: int
    cSecVal: str
    lSecVal: str


class _StaticThreshold(_TypedDict):
    insCode: None
    dEven: int
    hEven: int
    psGelStaMax: float
    psGelStaMin: float


class _InstrumentInfo(_TypedDict):
    eps: _Eps
    sector: _Sector
    staticThreshold: _StaticThreshold
    minWeek: float
    maxWeek: float
    minYear: float
    maxYear: float
    qTotTran5JAvg: float
    kAjCapValCpsIdx: str
    dEven: int
    topInst: int
    faraDesc: str
    contractSize: int
    nav: float
    underSupervision: int
    cValMne: None
    lVal18: str
    cSocCSAC: None
    lSoc30: None
    yMarNSC: None
    yVal: str
    insCode: str
    lVal30: str
    lVal18AFC: str
    flow: int
    cIsin: str
    zTitad: float
    baseVol: int
    instrumentID: str
    cgrValCot: str
    cComVal: str
    lastDate: int
    sourceID: int
    flowTitle: str
    cgrValCotTitle: str


def _parse_market_state(s: str) -> _MarketState:
    datetime, tse_status, tse_index, tse_index_change \
        , tse_value, tse_tvol, tse_tval, tse_tno \
        , fb_status, fb_tvol, fb_tval, fb_tno \
        , derivatives_status, derivatives_tvol, derivatives_tval, derivatives_tno \
        , _ = s.split(',')
    if tse_index_change:  # can be '' before market start
        index_change_match = _INDEX_CHANGE_MATCH(tse_index_change)
        tse_index_change = float(index_change_match[2])
        if index_change_match[1] is not None:  # negative value
            tse_index_change *= -1
        if (m3 := index_change_match[3]) is not None:
            tse_index_change_percent  = float(m3)
        else:
            tse_index_change_percent = None
    else:
        tse_index_change_percent = None
    timestamp_match = _INDEX_TIMESTAMP_MATCH(datetime)
    result = {
        'datetime': _jdatetime(
            1400 + int(timestamp_match[1]), int(timestamp_match[2])
            , int(timestamp_match[3]), int(timestamp_match[4])
            , int(timestamp_match[5]), int(timestamp_match[6]))
        , 'tse_status': tse_status
        , 'tse_index': float(tse_index)
        , 'tse_index_change': tse_index_change
        , 'tse_tvol': float(tse_tvol)
        , 'tse_tval': float(tse_tval)
        , 'tse_tno': float(tse_tval)
        , 'fb_status': fb_status
        , 'fb_tvol': float(fb_tvol)
        , 'fb_tval': float(fb_tval)
        , 'fb_tno': int(fb_tno)
        , 'derivatives_status': derivatives_status
        , 'derivatives_tvol': float(derivatives_tvol)
        , 'derivatives_tval': float(derivatives_tval)
        , 'derivatives_tno': int(derivatives_tno)
        , 'tse_index_change_percent': tse_index_change_percent
        , 'tse_value': float(tse_value) if tse_value else None
    }
    return result


def _parse_ombud_messages(text) -> _DataFrame:
    headers = _findall(r'<th>(.+?)</th>', text)
    dates = _findall(r"<th class='ltr'>(.+?)</th>", text)
    descriptions = _findall(r'<td colspan="2">(.+?)<hr />\s*</td>', text)
    df = _DataFrame({
        'header': headers, 'date': dates, 'description': descriptions
    }, dtype='string', copy=False)
    if dates:  # pandas cannot do ('14' + df['date']) on empty dates
        df['date'] = ('14' + df['date']).apply(
            _jstrptime, format='%Y/%m/%d %H:%M')
    else:
        df['date'] = df['date'].astype(object)
    return df


SESSION : _ClientSession | None = None


class Session:
    """Create and return a ClientSession object.

    Use
    ``ClientTimeout(total=30., sock_connect=5., sock_read=5.)``
    as the default timeout.
    """

    def __new__(cls, *args, **kwargs) -> _ClientSession:
        global SESSION
        if 'timeout' not in kwargs:
            kwargs['timeout'] = _ClientTimeout(total=30., sock_connect=30., sock_read=30.)
        SESSION = _ClientSession(**kwargs)
        return SESSION


_FARSI_NORM = ''.maketrans('يك', 'یک')
_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:100.0) Gecko/20100101 Firefox/100.0'
}


# this function should only be called from _get below
async def _session_get(url: str) -> bytes:
    r = await SESSION.get(url, headers=_HEADERS)
    if r.history:
        _warn(f'r.history is not empty (possible redirection): {r.history}')
    return await r.read()


async def _get(url: str, *, fa=False) -> str | bytes:
    content = await _session_get(url)
    if fa is True:
        return content.decode().translate(_FARSI_NORM)
    return content


_DOMAIN = 'http://old.tsetmc.com/'
# API does not work on www domain
_API = 'http://cdn.tsetmc.com/api/'


async def _get_data(path: str, *, fa=False) -> str | bytes:
    return await _get(f'{_DOMAIN}tsev2/data/' + path, fa=fa)


async def _get_par_tree(path: str, *, fa=True) -> str | bytes:
    return await _get(f'{_DOMAIN}Loader.aspx?ParTree={path}', fa=fa)


async def _api(path: str, *, fa=False):
    content = await _get(f'{_API}{path}', fa=fa)
    try:
        return loads(content)
    except JSONDecodeError:
        error(f'url={_API}{path}\n{content=}')
        raise


def _numerize(df: _DataFrame, cols: tuple[str, ...], astype=float, comma=False):
    for col in cols:
        c = df[col]
        if comma is True:
            c = c.str.replace(',', '')
        # https://stackoverflow.com/a/39684629/2705757
        df[col] = (
            c.replace(r' [KMB]$', '', regex=True).astype(astype)
        ) * c.str.extract(r'[\d\.]+([KMBT]+)', expand=False).fillna(1).replace(
            ['K', 'M', 'B', 'T'], [10 ** 3, 10 ** 6, 10 ** 9, 10 ** 12]
        )



from datetime import datetime
from types import NoneType
from unittest.mock import patch

from aiohttp_test_utils import file, files
from jdatetime import datetime as jdatetime
from numpy import dtype, int64
from pandas import DataFrame, DatetimeIndex
from pytest import raises, warns

from tests import assert_dict_type, assert_market_state
from tsetmc import _InstrumentInfo

# noinspection PyProtectedMember
from tsetmc.instruments import (
    _ETF,
    Instrument,
    _ClientType,
    _ClosingPriceInfo,
    _Codal,
    _Identity,
    _LiveData,
    _Message,
    _parse_price_info,
    _Search,
    _ShareHolder,
    _ShareHolderCompany,
    old_search,
    price_adjustments,
    search,
    share_holder_companies,
)


def assert_page_data(
    d, general=True, trade_history=False, related_companies=False
):
    if trade_history:
        trade_history = d.pop('trade_history')
        assert [*trade_history.dtypes.items()] == [
            ('pc', dtype('float64')),
            ('py', dtype('float64')),
            ('pmin', dtype('float64')),
            ('pmax', dtype('float64')),
            ('tno', dtype('int64')),
            ('tvol', dtype('int64')),
            ('tval', dtype('float64')),
        ]
        assert trade_history.index.dtype == dtype('<M8[ns]')
        assert trade_history.index.name == 'date'

    if related_companies:
        related_companies = d.pop('related_companies')
        assert type(related_companies) is list
        assert type(related_companies[0]) is Instrument

    if general:
        assert type(d.pop('sps')) in (float, NoneType)
        for k in ('eps', 'free_float'):
            assert type(d.pop(k)) in (int, NoneType)
        for k in ('bvol', 'cs', 'flow', 'month_average_volume', 'z'):
            assert type(d.pop(k)) is int
        for k in (
            'sector_pe',
            'tmax',
            'tmin',
            'week_max',
            'week_min',
            'year_max',
            'year_min',
        ):
            assert type(d.pop(k)) is float
        assert d.keys() == {
            'cisin',
            'flow_name',
            'isin',
            'group_code',
            'l18',
            'l30',
            'sector_name',
        }
        assert all(type(v) is str for v in d.values())


@file('fmelli.html')
async def test_page_data():
    ins = Instrument(35425587644337450)
    assert ins._l18 is ins._l30 is None
    with warns(DeprecationWarning):
        d = await ins.page_data(True, True, True)
    assert_page_data(d, True, True, True)


@file('dey.html')
async def test_page_data_no_free_float():
    with warns(DeprecationWarning):
        d = await Instrument(44818950263583523).page_data()
    assert_page_data(d)


@file('kala.html')
async def test_page_data_no_eps():
    with warns(DeprecationWarning):
        d = await Instrument(44549439964296944).page_data()
    assert_page_data(d)


@file('khgostar.html')
async def test_page_data_negative_sector_pe():
    with warns(DeprecationWarning):
        d = await Instrument(48990026850202503).page_data()
    assert_page_data(d)


def assert_live_data(
    d: _LiveData, best_limits=False, market_state=False, nav=False
):
    if best_limits:
        best_limits = d.pop('best_limits')
        assert [*best_limits.dtypes.items()] == [
            ('zd', dtype('int64')),
            ('qd', dtype('int64')),
            ('pd', dtype('int64')),
            ('po', dtype('int64')),
            ('qo', dtype('int64')),
            ('zo', dtype('int64')),
        ]

    if market_state:
        market_state = d.pop('market_state', None)
        if market_state is not None:
            assert_market_state(market_state)

    if nav:
        assert type(d.pop('nav_datetime')) in (jdatetime, str)
        assert type(d.pop('nav')) is int

    assert type(d.pop('datetime')) is datetime
    for k in ('timestamp', 'status'):
        assert type(d.pop(k)) is str
    assert [*d.keys()] == [
        'pl',
        'pc',
        'pf',
        'py',
        'pmin',
        'pmax',
        'tno',
        'tvol',
        'tval',
    ]
    assert all(type(v) is int for v in d.values())


@file('dara_yekom.txt')
async def test_dara1_instant():
    with warns(DeprecationWarning):
        d = await Instrument(62235397452612911).live_data(
            market_state=True, best_limits=True
        )
    assert_live_data(d, best_limits=True, market_state=True, nav=True)


@file('asam.txt')
async def test_asam_instant():
    with warns(DeprecationWarning):
        d = await Instrument(36592972482259020).live_data(best_limits=True)
    assert_live_data(d, best_limits=True, nav=True)


@file('negin.txt')
async def test_negin_instant():
    with warns(DeprecationWarning):
        d = await Instrument(10145129193828624).live_data()
    assert_live_data(d, nav=True)


FMELLI = Instrument(35425587644337450)


@file('fmelli.txt')
async def test_fmelli_instant():
    with warns(DeprecationWarning):
        d = await FMELLI.live_data(best_limits=False)
    assert_live_data(d, best_limits=False)


@file('vskhooz_short_response.txt')
async def test_vskhooz_short():
    with warns(DeprecationWarning):
        assert_live_data(await Instrument(5454781314262062).live_data())


@file('vskhooz_long_response.txt')
async def test_vskhooz_long():
    with warns(DeprecationWarning):
        assert_live_data(await Instrument(5454781314262062).live_data())


@file('fmelli_trade_history_top2.txt')
async def test_trade_history():
    with warns(DeprecationWarning):
        df0 = await FMELLI.trade_history(2)
    assert [*df0.dtypes.items()] == [
        ('pmax', dtype('float64')),
        ('pmin', dtype('float64')),
        ('pc', dtype('float64')),
        ('pl', dtype('float64')),
        ('pf', dtype('float64')),
        ('py', dtype('float64')),
        ('tval', dtype('float64')),
        ('tvol', dtype('int64')),
        ('tno', dtype('int64')),
    ]
    assert df0.index.name == 'date'
    assert isinstance(df0.index, DatetimeIndex)
    with warns(DeprecationWarning):
        df1 = await FMELLI.trade_history(2, 1)
    assert len(df1) >= len(df0)


VSADID = Instrument('41713045190742691')


@file('vsadid.txt')
async def test_vsadid():
    with warns(DeprecationWarning):
        d = await VSADID.live_data()
    assert_live_data(d)


@file('search_firuze.json')
async def test_from_search_with_numeric_description():
    # note the "30" in فيروزه - صندوق شاخص30 شركت فيروزه- سهام
    i = await Instrument.from_search('فیروزه')
    assert i.code == '66036975502302203'
    assert await i.l18 == 'فیروزه'


async def test_repr():
    # known l18
    assert (
        repr(await Instrument.from_l18('فملی'))
        == "Instrument(35425587644337450, 'فملی')"
    )
    # unknown l18
    assert repr(Instrument(1)) == "Instrument(1)"
    assert (
        repr(Instrument(1, l30='مجتمع جهان فولاد سيرجان'))
        == "Instrument(1, l30='مجتمع جهان فولاد سيرجان')"
    )


async def test_equal():
    assert await Instrument.from_l18('فملی') == Instrument(35425587644337450)


@file('vsadid_identification.html')
async def test_identification():
    with warns(DeprecationWarning):
        identification = await VSADID.identification()
    assert identification == {
        'بازار': 'بازار پایه زرد فرابورس',
        'زیر گروه صنعت': 'استخراج سایر فلزات اساسی',
        'نام شرکت': 'گروه\u200cصنعتی\u200cسدید',
        'نام لاتین شرکت': 'Sadid Group',
        'نماد 30 رقمی فارسی': 'گروه \u200cصنعتی\u200cسدید',
        'نماد فارسی': 'وسدید - لغو پذیرش شده',
        'کد 12 رقمی شرکت': 'IRO7SDIP0002',
        'کد 12 رقمی نماد': 'IRO7SDIP0001',
        'کد 4 رقمی شرکت': 'SDIP',
        'کد 5 رقمی نماد': 'SDIP1',
        'کد تابلو': '7',
        'کد زیر گروه صنعت': '2799',
        'کد گروه صنعت': '27',
        'گروه صنعت': 'فلزات اساسی',
    }


@file('opal_client_types.txt')
async def test_client_type_history_old():
    with warns(DeprecationWarning):
        df = await Instrument(655060129740445).client_type_history_old()
    assert [*df.dtypes.items()] == [
        ('n_buy_count', dtype('int64')),
        ('l_buy_count', dtype('int64')),
        ('n_sell_count', dtype('int64')),
        ('l_sell_count', dtype('int64')),
        ('n_buy_volume', dtype('int64')),
        ('l_buy_volume', dtype('int64')),
        ('n_sell_volume', dtype('int64')),
        ('l_sell_volume', dtype('int64')),
        ('n_buy_value', dtype('int64')),
        ('l_buy_value', dtype('int64')),
        ('n_sell_value', dtype('int64')),
        ('l_sell_value', dtype('int64')),
    ]


@file('faraz_GetClientTypeHistory_20220222.json')
async def test_client_type_history():
    d = await Instrument(13666407494621646).client_type_history(20220222)
    assert d.keys() == {
        'recDate',
        'insCode',
        'buy_I_Volume',
        'buy_N_Volume',
        'buy_I_Value',
        'buy_N_Value',
        'buy_N_Count',
        'sell_I_Volume',
        'buy_I_Count',
        'sell_N_Volume',
        'sell_I_Value',
        'sell_N_Value',
        'sell_N_Count',
        'sell_I_Count',
    }


@file('fzar_GetClientTypeHistory_all.json')
async def test_client_type_history_no_date():
    df = await Instrument(8175784894140974).client_type_history()
    assert [*df.dtypes.items()] == [
        ('recDate', dtype('int64')),
        ('insCode', dtype('O')),
        ('buy_I_Volume', dtype('float64')),
        ('buy_N_Volume', dtype('float64')),
        ('buy_I_Value', dtype('float64')),
        ('buy_N_Value', dtype('float64')),
        ('buy_N_Count', dtype('int64')),
        ('sell_I_Volume', dtype('float64')),
        ('buy_I_Count', dtype('float64')),
        ('sell_N_Volume', dtype('float64')),
        ('sell_I_Value', dtype('float64')),
        ('sell_N_Value', dtype('float64')),
        ('sell_N_Count', dtype('int64')),
        ('sell_I_Count', dtype('int64')),
    ]


AVA = Instrument('18007109712724189')


@files(  # id_cisin may change from time to time
    'ava_holders.txt',
    'ava_holder.txt',
    'ava_holder.txt',
    'ava_holder.txt',
    'ava_holder.txt',
)
async def test_holders_holder():
    with warns(DeprecationWarning):
        holders = await AVA.holders(cisin='IRT3AVAF0003')
    assert [*holders.dtypes.items()] == [
        ('holder', dtype('O')),
        ('shares/units', dtype('O')),
        ('%', dtype('float64')),
        ('change', dtype('int64')),
        ('id_cisin', dtype('O')),
    ]

    id_cisin = holders.at[0, 'id_cisin']

    with warns(DeprecationWarning):
        hist, oth = await AVA.holder(id_cisin, True, True)
    assert [*hist.dtypes.items()] == [('shares', dtype('int64'))]
    assert oth.index.name == 'ins_code'
    assert hist.index.dtype.kind == 'M'
    assert [*oth.dtypes.items()] == [
        ('name', dtype('O')),
        ('shares', dtype('int64')),
        ('percent', dtype('float64')),
    ]
    with warns(DeprecationWarning):
        hist = await AVA.holder('43789,IRT3AVAF0003', True)
    assert type(hist) is DataFrame
    with warns(DeprecationWarning):
        oth = await AVA.holder('43789,IRT3AVAF0003', False, True)
    assert type(oth) is DataFrame
    with warns(DeprecationWarning):
        result = await AVA.holder('43789,IRT3AVAF0003', False)
    assert oth.equals(result)


@files(  # share_holder_share_id may change from time to time
    'ava_holders.json',
    'share_holder_companies.json',
    'ava_share_holder_history.json',
)
async def test_share_holders_companies_histories():
    holders = await AVA.share_holders()
    first_holder = holders[0]
    assert_dict_type(first_holder, _ShareHolder)

    share_holder_share_id = first_holder['shareHolderShareID']

    companies = await share_holder_companies(share_holder_share_id)
    assert_dict_type(companies[0], _ShareHolderCompany)

    df = await AVA.share_holder_history(
        share_holder_id=share_holder_share_id,
        days=2,
    )
    assert len(df) == 2
    assert [*df.dtypes.items()] == [
        ('shareHolderID', dtype('int64')),
        ('shareHolderName', dtype('O')),
        ('cIsin', dtype('O')),
        ('numberOfShares', dtype('float64')),
        ('perOfShares', dtype('float64')),
        ('change', dtype('int64')),
        ('changeAmount', dtype('float64')),
        ('shareHolderShareID', dtype('int64')),
    ]
    assert df.index.dtype == dtype('<M8[ns]')


@file('vsadid_identification.html')
async def test_holders_without_cisin():
    inst = VSADID
    with warns(DeprecationWarning):
        d = await inst.identification()
    assert d['کد 12 رقمی شرکت'] == 'IRO7SDIP0002'
    with patch.object(
        Instrument, 'info', side_effect=NotImplementedError
    ) as info:
        with raises(NotImplementedError), warns(DeprecationWarning):
            await inst.holders()
    info.assert_called_once()


@file('fmelli_price_adjustment.html')
async def test_adjustments():
    with warns(DeprecationWarning):
        df = await FMELLI.adjustments()
    assert len(df) >= 18
    assert [*df.dtypes.items()] == [
        ('date', dtype('O')),
        ('adj_pc', dtype('int64')),
        ('pc', dtype('int64')),
    ]
    assert type(df.iat[0, 0]) is jdatetime


@file('price_adjustment_api.json')
async def test_price_adjustments_method():
    df = await FMELLI.price_adjustments()
    assert len(df) >= 18
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('pClosingNotAdjusted', dtype('float64')),
        ('corporateTypeCode', dtype('O')),
        ('instrument', dtype('O')),
    ]
    assert df.index[-1] == datetime(2009, 7, 15)


@file('adjustments_flow_7.html')
async def test_price_adjustments():
    df = await price_adjustments(7)
    assert [*df.dtypes.items()] == [
        ('l18', dtype('O')),
        ('l30', dtype('O')),
        ('date', dtype('O')),
        ('adj_pc', dtype('int64')),
        ('pc', dtype('int64')),
    ]
    assert len(df) == 6
    assert df.iat[-1, -1] == 1000


@file('latif_financial_aph.aspx')
async def test_adjusted_price_history():
    df = await Instrument(16422980660132735).price_history()
    assert df.index.name == 'date'
    assert df.index.dtype == dtype('<M8[ns]')
    assert [*df.dtypes.items()] == [
        ('pmax', dtype('int64')),
        ('pmin', dtype('int64')),
        ('pf', dtype('int64')),
        ('pl', dtype('int64')),
        ('tvol', dtype('int64')),
        ('pc', dtype('int64')),
    ]


@file('search_mellat.txt')
async def test_old_search():
    df = await old_search('ملت')
    assert type(df) is DataFrame
    assert len(df) > 40
    assert [*df.dtypes.items()] == [
        ('l18', dtype('O')),
        ('l30', dtype('O')),
        ('ins_code', dtype('int64')),
        ('retail', dtype('int64')),
        ('compensation', dtype('int64')),
        ('wholesale', dtype('int64')),
        ('_unknown1', dtype('int64')),
        ('_unknown2', dtype('int64')),
        ('_unknown3', dtype('int64')),
        ('_unknown4', dtype('int64')),
        ('_unknown5', dtype('O')),
    ]


@file('search_atlas.txt')
async def test_search():
    r = await search('اطلس')
    assert type(r) is list
    assert_dict_type(r[0], _Search)


async def test_l18_without_web_request():
    assert await Instrument(46348559193224090).l18 == 'فولاد'


@file('fmelli_introduction.html')
async def test_introduction():
    with warns(DeprecationWarning):
        d = await (await Instrument.from_l18('فملی')).introduction()
    assert d == {
        'موضوع فعالیت': 'اکتشافات، استخراج و بهره برداری از معادن مس ایران',
        'مدیر عامل': 'علی رستمی',
        'نشانی': 'مجتمع مس سرچشمه و مجتمع مس میدوک در استان کرمان و مجتمع مس سونگون در تبریز شهرستان ورزقان واقع شده اند.',
        'شماره تلفن': '021-88724410',
        'نمابر': '021-88729014',
        'نشانی دفتر': 'تهران خیابان ولیعصر نرسیده به پارک ساعی پلاک 2161',
        'نشانی امور سهام': 'تهران خیابان خالد اسلامبولی (وزراء) خیابان 11 پلاک 22 طبقه اول',
        'وب سایت': 'www.nicico.com',
        'ایمیل': 'office@nicico.com',
        'حسابرس': 'حسابرسی وخدمات مالی ومدیریت ایران مشهود',
        'سرمایه': '400000000',
        'سال مالی': '12/29',
        'مدیر مالی': 'امید علی معیری',
        'شناسه ملی': '10100582059',
        'شرکت مدیریت فناوری بورس تهران فقط گردآورنده ی اطلاعات برای معرفی شرکت های پذیرفته شده در بورس و فرابورس است. این شرکت نظارت،مسئولیت یا آگاهی در باره ی چگونگی ارائه خدمات اداره سهام شرکت ها، توزیع سود، وضعیت افزایش سرمایه آنها و موارد مشابه دیگر ندارد و مرجع رسیدگی به شکایات یا رفع کاستی های شرکت ها نمی باشد.': 'شرکت مدیریت فناوری بورس تهران فقط گردآورنده ی اطلاعات برای معرفی شرکت های پذیرفته شده در بورس و فرابورس است. این شرکت نظارت،مسئولیت یا آگاهی در باره ی چگونگی ارائه خدمات اداره سهام شرکت ها، توزیع سود، وضعیت افزایش سرمایه آنها و موارد مشابه دیگر ندارد و مرجع رسیدگی به شکایات یا رفع کاستی های شرکت ها نمی باشد.',
        'برای مشاهده اطلاعات دقیق به کدال مراجعه کنید.': 'برای مشاهده اطلاعات دقیق به کدال مراجعه کنید.',
    }


@file('fmelli_publisher.json')
async def test_publisher():
    inst = await Instrument.from_l18('فملی')
    publisher = await inst.publisher()
    assert publisher.keys() == {
        'activitySubject',
        'address',
        'auditorName',
        'companyId',
        'companyType',
        'companyType1',
        'displaySymbol',
        'email',
        'enActivitySubject',
        'enAddress',
        'enDisplayedSymbol',
        'enExecutiveManager',
        'enFinancialManager',
        'enInspector',
        'enManagementGroup',
        'enName',
        'enOfficeAddress',
        'enShareOfficeAddress',
        'executiveManager',
        'faxNo',
        'financialManager',
        'financialYear',
        'id',
        'inspListedCapitalector',
        'inspector',
        'isic',
        'listedCapital',
        'managementGroup',
        'name',
        'nationalCode',
        'officeAddress',
        'reportingType',
        'shareOfficeAddress',
        'state',
        'stateName',
        'symbol',
        'telNo',
        'website',
    }


@file('tajalli_ombud_messages.html')
async def test_ombud_messages():
    with warns(DeprecationWarning):
        df = await Instrument(1301069819790264).ombud_messages()
    assert [*df.dtypes.items()] == [
        ('header', 'string[python]'),
        ('date', dtype('O')),
        ('description', 'string[python]'),
    ]
    assert type(df.iat[0, 1]) is jdatetime


@file('shetab_messages.json')
async def test_messages():
    messages = await Instrument(64216772923447100).messages()
    assert_dict_type(messages[0], _Message)


@file('fmelli_dps.txt')
async def test_dps_history():
    df = await FMELLI.dps_history()
    assert [*df.dtypes.items()] == [
        ('publish_date', dtype('O')),
        ('meeting_date', dtype('O')),
        ('fiscal_year', dtype('O')),
        ('profit_or_loss_after_tax', dtype('float64')),
        ('distributable_profit', dtype('float64')),
        ('accumulated_profit_at_the_end_of_the_period', dtype('float64')),
        ('cash_earnings_per_share', dtype('float64')),
    ]
    assert type(df.iat[0, 0]) is jdatetime
    assert type(df.iat[0, 1]) is jdatetime
    assert type(df.iat[0, 2]) is jdatetime


def test_hash():
    assert hash(Instrument(int64(1))) == 1


def test_parse_price_info():
    assert _parse_price_info(  # len == 17
        '12:29:23,A ,75910,75850,76030,75860,76280,75500,887,1599031,121280988420,1,20220430,122923,121280988420,1401/2/10 15:30:00,79184'
    ) == _parse_price_info(  # len == 16
        '12:29:23,A ,75910,75850,76030,75860,76280,75500,887,1599031,121280988420,0,20220430,122923,1401/2/10 15:30:00,79184'
    )


def test_parse_price_info_bad_date():
    # 1400/12/30 is not a valid date
    assert (
        _parse_price_info(
            '12:19:35,A ,63070,64270,62540,63550,64300,62540,30,7973,512457280,1,20220813,121935,1400/12/30 16:30:00,63354'
        )['nav_datetime']
        == '1400/12/30 16:30:00'
    )


KARIS = Instrument(69067576215760005)


@file('karis_info.json')
async def test_info_on_etf():
    info = await KARIS.info()
    assert_dict_type(info, _InstrumentInfo)


@file('fmelli_info.json')
async def test_info_on_fmelli():
    info = await FMELLI.info()
    assert_dict_type(info, _InstrumentInfo)


@file('karis_trades.json')
async def test_trades():
    df = await KARIS.trades()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('nTran', dtype('int64')),
        ('hEven', dtype('int64')),
        ('qTitTran', dtype('int64')),
        ('pTran', dtype('float64')),
        ('qTitNgJ', dtype('int64')),
        ('iSensVarP', dtype('O')),
        ('pPhSeaCotJ', dtype('float64')),
        ('pPbSeaCotJ', dtype('float64')),
        ('iAnuTran', dtype('int64')),
        ('xqVarPJDrPRf', dtype('float64')),
        ('canceled', dtype('int64')),
    ]


@file('karis_codal.json')
async def test_codal():
    d = await KARIS.codal(n=3)
    assert len(d) == 3
    assert_dict_type(d[0], _Codal)


@file('daily_closing_price_karis.json')
async def test_daily_closing_price():
    df = await KARIS.daily_closing_price(n=3)
    assert [*df.dtypes.items()] == [
        ('priceChange', dtype('float64')),
        ('priceMin', dtype('float64')),
        ('priceMax', dtype('float64')),
        ('priceYesterday', dtype('float64')),
        ('priceFirst', dtype('float64')),
        ('last', dtype('bool')),
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
    ]
    assert len(df) == 3
    assert df.index.dtype == dtype('<M8[ns]')


@file('closing_price_info_karis.json')
async def test_closing_price_info():
    info = await KARIS.closing_price_info()
    assert_dict_type(info, _ClosingPriceInfo)


@file('best_limits.json')
async def test_best_limits():
    df = await KARIS.best_limits()
    assert [*df.dtypes.items()] == [
        ('number', dtype('int64')),
        ('qTitMeDem', dtype('int64')),
        ('zOrdMeDem', dtype('int64')),
        ('pMeDem', dtype('float64')),
        ('pMeOf', dtype('float64')),
        ('zOrdMeOf', dtype('int64')),
        ('qTitMeOf', dtype('int64')),
        ('insCode', dtype('O')),
    ]
    assert len(df) == 5


@file('client_type_karis.json')
async def test_client_type():
    d = await KARIS.client_type()
    assert_dict_type(d, _ClientType)


@file('etf_karis.json')
async def test_etf():
    d = await KARIS.etf()
    assert_dict_type(d, _ETF)


@file('related_companies_karis.json')
async def test_related_companies():
    # await KARIS.info()['sector']['cSecVal'] == "68 "
    d = await KARIS.related_companies(cs="68 ")
    c = d['relatedCompany']
    h = d['relatedCompanyThirtyDayHistory']

    assert [*c.dtypes.items()] == [
        ('instrumentState', dtype('O')),
        ('lastHEven', dtype('int64')),
        ('finalLastDate', dtype('int64')),
        ('nvt', dtype('float64')),
        ('mop', dtype('int64')),
        ('thirtyDayClosingHistory', dtype('O')),
        ('priceChange', dtype('float64')),
        ('priceMin', dtype('float64')),
        ('priceMax', dtype('float64')),
        ('priceYesterday', dtype('float64')),
        ('priceFirst', dtype('float64')),
        ('last', dtype('bool')),
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
        ('instrument.cValMne', dtype('O')),
        ('instrument.lVal18', dtype('O')),
        ('instrument.cSocCSAC', dtype('O')),
        ('instrument.lSoc30', dtype('O')),
        ('instrument.yMarNSC', dtype('O')),
        ('instrument.yVal', dtype('O')),
        ('instrument.insCode', dtype('O')),
        ('instrument.lVal30', dtype('O')),
        ('instrument.lVal18AFC', dtype('O')),
        ('instrument.flow', dtype('int64')),
        ('instrument.cIsin', dtype('O')),
        ('instrument.zTitad', dtype('float64')),
        ('instrument.baseVol', dtype('int64')),
        ('instrument.instrumentID', dtype('O')),
        ('instrument.cgrValCot', dtype('O')),
        ('instrument.cComVal', dtype('O')),
        ('instrument.lastDate', dtype('int64')),
        ('instrument.sourceID', dtype('int64')),
        ('instrument.flowTitle', dtype('O')),
        ('instrument.cgrValCotTitle', dtype('O')),
    ]
    assert [*h.dtypes.items()] == [
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
    ]

    assert h.groupby('insCode')['insCode'].agg(len).mode()[0] == 29


@file('test_identity.json')
async def test_identity():
    d = await KARIS.identity()
    assert_dict_type(d, _Identity)



"""http://redirectcdn.tsetmc.com/Site.aspx?ParTree=11141111"""
from pandas import read_html as _read_html

from tsetmc import _DOMAIN, _get
from tsetmc.general import _make_soup


async def _site_partree(params: str):
    return await _get(
        f'{_DOMAIN}Site.aspx?ParTree={params}',
        fa=True,
    )


async def client_type() -> dict:
    """http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111116&LnkIdn=3568"""
    text = await _site_partree('1114111116&LnkIdn=3568')
    out = _read_html(text)[2]
    out.rename(columns=out.iloc[0], inplace=True)
    return {
        'output': out,
    }


async def instrument_filter_by_date() -> dict:
    """http://redirectcdn.tsetmc.com/Site.aspx?ParTree=111411111Z&LnkIdn=834"""
    text = await _site_partree('1114111116&LnkIdn=834')
    soup = _make_soup(text)

    ul0, ul1, ul2 = soup.select('td > ul')

    flow = {int(k): v for k, v in [i.text.split(' : ') for i in ul0.select('li')]}
    yval = {k: v for k, v in [i.text.split(': ') for i in ul1.select('li')]}

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'Flow': flow,
            'YVal': yval,
        },
    }


async def instrument_state() -> dict:
    """http://redirectcdn.tsetmc.com/Site.aspx?ParTree=111411111Y&LnkIdn=833"""
    text = await _site_partree('1114111116&LnkIdn=833')
    soup = _make_soup(text)

    tds = soup.select('td')
    flow = {int(k): v for k, v in [i.text.strip().split(' : ') for i in tds[7].select('li')]}
    cetaval = {k.strip('"'): v.strip('"') for k, v in [i.text.strip().split(':') for i in tds[-1].select('p')]}

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'CEtaVal': cetaval,
        },
    }


async def instrument() -> dict:
    """http://tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83"""
    text = await _site_partree('1114111116&LnkIdn=83')
    soup = _make_soup(text)

    t0 = soup.select_one('table')
    t1, t2, t3 = t0.select('table')

    flow = {int(k): v for k, v in [i.text.split(' : ') for i in t1.select('li')]}

    yval = {
        int(k.text): (v1.text.strip(), v2.text.strip())
        for k, v1, v2 in [tr.select('td') for tr in t3.select('tr')]
    }

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'YVal': yval,
        },
    }


async def best_limits_all_ins() -> dict:
    """http://redirectcdn.tsetmc.com/Site.aspx?ParTree=111411111P&LnkIdn=97"""
    text = await _site_partree('111411111P&LnkIdn=97')
    soup = _make_soup(text)

    t0 = soup.select_one('table')
    t1, t2 = t0.select('table')

    flow = {int(k): v for k, v in [i.text.split(' : ') for i in t1.select('li')]}
    out = {k.text.strip(): v.text.strip() for k, v in [tr.select('td') for tr in t2.select('tr')]}

    return {
        'input': {
            'Flow': flow,
        },
        'output': out,
    }


async def trade_last_day() -> dict:
    """http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111113&LnkIdn=84"""
    text = await _site_partree('1114111113&LnkIdn=84')
    soup = _make_soup(text)

    flow = {int(k): v for k, v in [i.text.split(' : ') for i in soup.select_one('ul').select('li')]}
    out = {k.text.strip(): v.text.strip() for k, v in [tr.select('td') for tr in soup.select('table:last-child tr')]}

    return {
        'input': {
            'Flow': flow,
        },
        'output': out,
    }



from aiohttp_test_utils import file
from jdatetime import datetime as jdatetime
from numpy import dtype
from pandas.api.types import is_numeric_dtype

from tests import assert_market_state

# noinspection PyProtectedMember
from tsetmc.market_watch import (
    _PRICE_DTYPES,
    _parse_market_state,
    client_type_all,
    closing_price_all,
    key_stats,
    market_watch_init,
    market_watch_plus,
    ombud_messages,
    status_changes,
)

PRICE_DTYPES_ITEMS = [*_PRICE_DTYPES.items()][1:]


@file('MarketWatchInit.aspx')
async def test_market_watch_init():
    mwi = await market_watch_init(join=False, market_state=False)
    assert [*mwi['prices'].dtypes.items()] == PRICE_DTYPES_ITEMS
    assert [*mwi['best_limits'].index.dtypes.items()] == [
        ('ins_code', 'string[python]'),
        ('number', dtype('int64')),
    ]
    assert 'market_state' not in mwi

    mwi = await market_watch_init(market_state=True)
    prices = mwi['prices']
    assert 'market_state' in mwi
    assert 'best_limits' in mwi
    assert [*prices.dtypes.items()] == [
        ('pd1', dtype('int64')),
        ('po1', dtype('int64')),
        ('qd1', dtype('int64')),
        ('qo1', dtype('int64')),
        ('zd1', dtype('int64')),
        ('zo1', dtype('int64')),
        ('pd2', dtype('int64')),
        ('po2', dtype('int64')),
        ('qd2', dtype('int64')),
        ('qo2', dtype('int64')),
        ('zd2', dtype('int64')),
        ('zo2', dtype('int64')),
        ('pd3', dtype('int64')),
        ('po3', dtype('int64')),
        ('qd3', dtype('int64')),
        ('qo3', dtype('int64')),
        ('zd3', dtype('int64')),
        ('zo3', dtype('int64')),
        ('pd4', dtype('int64')),
        ('po4', dtype('int64')),
        ('qd4', dtype('int64')),
        ('qo4', dtype('int64')),
        ('zd4', dtype('int64')),
        ('zo4', dtype('int64')),
        ('pd5', dtype('int64')),
        ('po5', dtype('int64')),
        ('qd5', dtype('int64')),
        ('qo5', dtype('int64')),
        ('zd5', dtype('int64')),
        ('zo5', dtype('int64')),
        *PRICE_DTYPES_ITEMS,
    ]

    i = prices.index
    assert i.name == 'ins_code'
    assert i.dtype == 'string[python]'

    mwi = await market_watch_init(prices=False, market_state=False)
    assert 'prices' not in mwi
    assert [*mwi['best_limits'].index.dtypes.items()] == [
        ('ins_code', 'string[python]'),
        ('number', dtype('int64')),
    ]


@file('ClosingPriceAll.aspx')
async def test_closing_price_all():
    df = await closing_price_all()
    assert all(t == 'int64' for t in df.dtypes)
    assert df.columns.to_list() == [
        'pc',
        'pl',
        'tno',
        'tvol',
        'tval',
        'pmin',
        'pmax',
        'py',
        'pf',
    ]
    assert [*df.index.dtypes.items()] == [
        ('ins_code', 'string[python]'),
        ('n', dtype('int64')),
    ]


@file('ClientTypeAll.aspx')
async def test_client_type_all():
    df = await client_type_all()
    assert all(
        df.columns
        == [
            'n_buy_count',
            'l_buy_count',
            'n_buy_volume',
            'l_buy_volume',
            'n_sell_count',
            'l_sell_count',
            'n_sell_volume',
            'l_sell_volume',
        ]
    )
    assert all(dt == 'int64' for dt in df.dtypes)
    assert df.index.name == 'ins_code'
    assert df.index.dtype == 'string[python]'


@file('InstValue.aspx')
async def test_key_stats():
    df = await key_stats()
    assert all(df.columns.str.startswith('is'))
    assert all(t == 'float64' for t in df.dtypes)
    assert df.index.name == 'ins_code'
    assert df.index.dtype == 'string[python]'


def test_parse_index():
    # no tse_value
    assert _parse_market_state(
        "00/1/14 06:40:12,F,1294521.64,<div class='mn'>(8671.45)</div>,,0.00,0.00,0,C,0.00,0.00,0,C,0.00,0.00,0,"
    ) == {
        'datetime': jdatetime(1400, 1, 14, 6, 40, 12),
        'derivatives_status': 'C',
        'derivatives_tno': 0,
        'derivatives_tval': 0.0,
        'derivatives_tvol': 0.0,
        'fb_status': 'C',
        'fb_tno': 0,
        'fb_tval': 0.0,
        'fb_tvol': 0.0,
        'tse_index': 1294521.64,
        'tse_index_change': -8671.45,
        'tse_index_change_percent': None,
        'tse_status': 'F',
        'tse_tno': 0.0,
        'tse_tval': 0.0,
        'tse_tvol': 0.0,
        'tse_value': None,
    }

    assert {
        'derivatives_status': 'N',
        'derivatives_tno': 2621,
        'derivatives_tval': 31470939000.0,
        'derivatives_tvol': 101096.0,
        'datetime': jdatetime(1400, 12, 16, 15, 45, 46),
        'fb_status': 'N',
        'fb_tno': 342228,
        'fb_tval': 163701347122693.0,
        'fb_tvol': 1057924358.0,
        'tse_index': 1169760.86,
        'tse_index_change': -8155.9,
        'tse_index_change_percent': -0.69,
        'tse_status': 'F',
        'tse_tno': 34288551133025.0,
        'tse_tval': 34288551133025.0,
        'tse_tvol': 3055466451.0,
        'tse_value': 4.674381234630472e16,
    } == _parse_market_state(
        "00/12/16 15:45:46,F,1169760.86,<div class='mn'>(8155.90)</div> -0.69%,46743812346304720.00,3055466451.00,34288551133025.00,428601,N,1057924358.00,163701347122693.00,342228,N,101096.00,31470939000.00,2621,"
    )

    assert {
        'derivatives_status': 'F',
        'derivatives_tno': 5796,
        'derivatives_tval': 57759844000.0,
        'derivatives_tvol': 225866.0,
        'datetime': jdatetime(1400, 12, 24, 14, 39, 40),
        'fb_status': 'F',
        'fb_tno': 544617,
        'fb_tval': 128484547655014.0,
        'fb_tvol': 1577202926.0,
        'tse_index': 1245186.04,
        'tse_index_change': 15808.56,
        'tse_index_change_percent': 1.29,
        'tse_status': 'F',
        'tse_tno': 75828635544957.0,
        'tse_tval': 75828635544957.0,
        'tse_tvol': 9682732949.0,
        'tse_value': 4.973605456635374e16,
    } == _parse_market_state(
        "00/12/24 14:39:40,F,1245186.04,<div class='pn'>15808.56</div> 1.29%,49736054566353740.00,9682732949.00,75828635544957.00,830860,F,1577202926.00,128484547655014.00,544617,F,225866.00,57759844000.00,5796,"
    )


@file('MarketWatchInit2.aspx')
async def test_market_watch_init_non_int_tmin_tmax():
    # ins_code 12785301426418659
    # used to raise
    # TypeError: cannot safely cast non-equivalent float64 to int64
    await market_watch_init()


@file('MarketWatchPlus00.txt')
async def test_market_watch_plus_new():
    mwp = await market_watch_plus(0, 0, messages=False, market_state=False)
    new_prices = mwp['new_prices']
    assert [*new_prices.dtypes.items()] == PRICE_DTYPES_ITEMS
    i = new_prices.index
    assert i.name == 'ins_code'
    assert i.dtype == 'string[python]'
    best_limits = mwp['best_limits']
    assert all(t == 'int64' for t in best_limits.dtypes)
    assert best_limits.columns.to_list() == [
        'number',
        'zo',
        'zd',
        'pd',
        'po',
        'qd',
        'qo',
    ]
    index = best_limits.index
    assert index.name == 'ins_code'
    assert index.dtype == 'string[python]'
    assert 'messages' not in mwp
    assert 'market_state' not in mwp


@file('MarketWatchPlus_h64130_r9540883525.txt')
async def test_market_watch_plus_update():
    mwp = await market_watch_plus(64130, 9540883525)
    price_updates = mwp['price_updates']
    assert price_updates.columns.to_list() == [
        'heven',
        'pf',
        'pc',
        'pl',
        'tno',
        'tvol',
        'tval',
        'pmin',
        'pmax',
    ]
    assert all(is_numeric_dtype(c) for c in price_updates.dtypes)
    assert price_updates.index.dtype == 'string[python]'

    market_state = mwp.pop('market_state', None)
    if market_state is not None:
        assert_market_state(market_state)

    for m in mwp['messages']:
        assert type(m) is str
        assert m.isnumeric()

    best_limits = mwp['best_limits']
    assert best_limits.columns.to_list() == [
        'number',
        'zo',
        'zd',
        'pd',
        'po',
        'qd',
        'qo',
    ]
    assert all(is_numeric_dtype(c) for c in best_limits.dtypes)
    assert best_limits.index.name == 'ins_code'
    assert best_limits.index.dtype == 'string[python]'

    assert type(mwp['refid']) == int

    new_prices = mwp['new_prices']
    assert [*new_prices.dtypes.items()] == PRICE_DTYPES_ITEMS
    i = new_prices.index
    assert i.name == 'ins_code'
    assert i.dtype == 'string[python]'


@file('status_changes.html')
async def test_status_changes():
    df = await status_changes(3)
    assert len(df) == 3
    assert (*df.dtypes.items(),) == (
        ('نماد', dtype('O')),
        ('نام', dtype('O')),
        ('وضعیت جدید', dtype('O')),
        ('date', dtype('O')),
    )
    assert type(df.iat[0, 3]) is jdatetime


@file('ombud_messages.html')
async def test_ombud_messages():
    df = await ombud_messages(top=3)
    assert len(df) == 3
    assert (*df.dtypes.items(),) == (
        ('header', 'string[python]'),
        ('date', dtype('O')),
        ('description', 'string[python]'),
    )
    assert type(df.iat[0, 1]) is jdatetime


@file('empty_ombud_messages.html')
async def test_ombud_messages_empty():
    # `sh_date` cannot be used without `containing`
    df = await ombud_messages(top=1, sh_date='1400-11-02', flow=0)
    assert df.empty
    assert (*df.dtypes.items(),) == (
        ('header', 'string[python]'),
        ('date', dtype('O')),
        ('description', 'string[python]'),
    )



from aiohttp_test_utils import file
from numpy import dtype

from tests import assert_dict_type
from tsetmc import _InstrumentInfo
from tsetmc.indices import Index, last_state


@file('last_state.json')
async def test_last_state():
    df = await last_state()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('xDrNivJIdx004', dtype('float64')),
        ('xPhNivJIdx004', dtype('float64')),
        ('xPbNivJIdx004', dtype('float64')),
        ('xVarIdxJRfV', dtype('float64')),
        ('last', dtype('bool')),
        ('indexChange', dtype('float64')),
        ('lVal30', dtype('O')),
        ('c1', dtype('int64')),
        ('c2', dtype('int64')),
        ('c3', dtype('int64')),
        ('c4', dtype('int64')),
    ]


OVERALL = Index('32097828799138957')


@file('index_info.json')
async def test_info():
    info = await OVERALL.info()
    assert_dict_type(info, _InstrumentInfo)


@file('index_last_day_history.json')
async def test_last_day_history():
    df = await OVERALL.last_day_history()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('O')),
        ('xDrNivJIdx004', dtype('float64')),
        ('xPhNivJIdx004', dtype('float64')),
        ('xPbNivJIdx004', dtype('float64')),
        ('xVarIdxJRfV', dtype('float64')),
        ('last', dtype('bool')),
        ('indexChange', dtype('float64')),
        ('lVal30', dtype('O')),
        ('c1', dtype('int64')),
        ('c2', dtype('int64')),
        ('c3', dtype('int64')),
        ('c4', dtype('int64')),
    ]
    assert df.index.dtype == 'datetime64[ns]'


@file('index_history.json')
async def test_history():
    df = await OVERALL.history()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('int64')),
        ('xNivInuClMresIbs', dtype('float64')),
        ('xNivInuPbMresIbs', dtype('float64')),
        ('xNivInuPhMresIbs', dtype('float64')),
    ]
    assert df.index.dtype == 'datetime64[ns]'


ICT = Index('41867092385281437')  # only has two companies currently


@file('index_companies.json')
async def test_companies():
    d = await ICT.companies()
    companies = d['indexCompany']
    companies_history = d['relatedCompanyThirtyDayHistory']
    if companies.empty:  # before the market start
        return
    assert [*companies.dtypes.items()] == [
        ('instrumentState', dtype('O')),
        ('lastHEven', dtype('int64')),
        ('finalLastDate', dtype('int64')),
        ('nvt', dtype('float64')),
        ('mop', dtype('int64')),
        ('thirtyDayClosingHistory', dtype('O')),
        ('priceChange', dtype('float64')),
        ('priceMin', dtype('float64')),
        ('priceMax', dtype('float64')),
        ('priceYesterday', dtype('float64')),
        ('priceFirst', dtype('float64')),
        ('last', dtype('bool')),
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
        ('instrument.cValMne', dtype('O')),
        ('instrument.lVal18', dtype('O')),
        ('instrument.cSocCSAC', dtype('O')),
        ('instrument.lSoc30', dtype('O')),
        ('instrument.yMarNSC', dtype('O')),
        ('instrument.yVal', dtype('O')),
        ('instrument.insCode', dtype('O')),
        ('instrument.lVal30', dtype('O')),
        ('instrument.lVal18AFC', dtype('O')),
        ('instrument.flow', dtype('int64')),
        ('instrument.cIsin', dtype('O')),
        ('instrument.zTitad', dtype('float64')),
        ('instrument.baseVol', dtype('int64')),
        ('instrument.instrumentID', dtype('O')),
        ('instrument.cgrValCot', dtype('O')),
        ('instrument.cComVal', dtype('O')),
        ('instrument.lastDate', dtype('int64')),
        ('instrument.sourceID', dtype('int64')),
        ('instrument.flowTitle', dtype('O')),
        ('instrument.cgrValCotTitle', dtype('O')),
    ]
    assert [*companies_history.dtypes.items()] == [
        ('id', dtype('int64')),
        ('insCode', dtype('O')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('iClose', dtype('bool')),
        ('yClose', dtype('bool')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
    ]



from unittest.mock import patch

from tsetmc.dataset import add_instrument
from tsetmc.instruments import Instrument


async def test_adding_existing_insturment(caplog):
    fmelli = Instrument(35425587644337450)
    i = {
        'lVal18AFC': 'فملی',
        'lVal30': 'ملی\u200c صنایع\u200c مس\u200c ایران\u200c',
    }
    with (
        patch.object(Instrument, 'info', return_value=i) as info,
        patch('tsetmc.dataset._dump') as dump,
    ):
        await add_instrument(fmelli)
    info.assert_awaited_once()
    dump.assert_not_called()
    assert len(records := caplog.records) == 1
    r = records[0]
    assert r.message == "l18 = 'فملی' already exists in dataset"
    assert r.levelname == 'WARNING'



from aiohttp_test_utils import file
from numpy import dtype

from tsetmc.docs import (
    best_limits_all_ins,
    client_type,
    instrument,
    instrument_filter_by_date,
    instrument_state,
    trade_last_day,
)


@file('client_type.html')
async def test_client_type():
    d = await client_type()
    out = d['output']
    assert len(out) == 9
    assert [*out.dtypes.items()] == [
        ('InsCode', dtype('O')),
        ('کد داخلی نماد', dtype('O')),
        ('Instrument Unique Key', dtype('O'))]


@file('instrument_filter_by_date.html')
async def test_instrument_filter_by_date():
    d = await instrument_filter_by_date()
    assert d['output']['YVal']['068'] == 'شاخص'


@file('instrument_state.html')
async def test_instrument_state():
    d = await instrument_state()
    assert d['output']['CEtaVal']['I '] == 'ممنوع'


@file('instrument.html')
async def test_instrument():
    d = await instrument()
    assert d['output']['YVal'][69] == ('شاخص', 'شاخص فرابورس')


@file('best_limits_all_ins.html')
async def test_best_limits_all_ins():
    d = await best_limits_all_ins()
    assert d['output']['QTitMeOf'] == 'حجم عرضه'


@file('trade_last_day.html')
async def test_trade_last_day():
    d = await trade_last_day()
    assert d['output']['HEven'] == 'ساعت'



from aiohttp_test_utils import file
from numpy import dtype

from tests import assert_dict_type
from tsetmc.general import (
    _MarketOverview,
    boards,
    cs_codes,
    industrial_groups_overview,
    major_holders_activity,
    market_map_data,
    market_overview,
    top_industry_groups,
)


@file('boards.html')
async def test_boards():
    assert await boards() == {
        1: 'تابلو اصلی',
        3: 'تابلو فرعی',
        4: 'تابلو غیر رسمی',
        5: 'فهرست اولیه',
        6: 'تابلو شاخص',
        7: 'فهرست مشروط',
        8: 'ابزار مشتقه',
        9: 'بازار اوراق بدهی',
    }


@file('cs_codes.html')
async def test_cs_codes():
    assert await cs_codes() == {
        '01': 'زراعت و خدمات وابسته',
        '02': 'جنگلداری و ماهیگیری',
        '10': 'استخراج زغال سنگ',
        '11': 'استخراج نفت گاز و خدمات جنبی جز اکتشاف',
        '13': 'استخراج کانه های فلزی',
        '14': 'استخراج سایر معادن',
        '15': 'حذف شده- فرآورده\u200cهای غذایی و آشامیدنی',
        '17': 'منسوجات',
        '19': 'دباغی، پرداخت چرم و ساخت انواع پاپوش',
        '20': 'محصولات چوبی',
        '21': 'محصولات کاغذی',
        '22': 'انتشار، چاپ و تکثیر',
        '23': 'فراورده های نفتی، کک و سوخت هسته ای',
        '24': 'حذف شده-مواد و محصولات شیمیایی',
        '25': 'لاستیک و پلاستیک',
        '26': 'تولید محصولات کامپیوتری الکترونیکی ونوری',
        '27': 'فلزات اساسی',
        '28': 'ساخت محصولات فلزی',
        '29': 'ماشین آلات و تجهیزات',
        '31': 'ماشین آلات و دستگاه\u200cهای برقی',
        '32': 'ساخت دستگاه\u200cها و وسایل ارتباطی',
        '33': 'ابزارپزشکی، اپتیکی و اندازه\u200cگیری',
        '34': 'خودرو و ساخت قطعات',
        '35': 'سایر تجهیزات حمل و نقل',
        '36': 'مبلمان و مصنوعات دیگر',
        '38': 'قند و شکر',
        '39': 'شرکتهای چند رشته ای صنعتی',
        '40': 'عرضه برق، گاز، بخاروآب گرم',
        '41': 'جمع آوری، تصفیه و توزیع آب',
        '42': 'محصولات غذایی و آشامیدنی به جز قند و شکر',
        '43': 'مواد و محصولات دارویی',
        '44': 'محصولات شیمیایی',
        '45': 'پیمانکاری صنعتی',
        '46': 'تجارت عمده فروشی به جز وسایل نقلیه موتور',
        '47': 'خرده فروشی،باستثنای وسایل نقلیه موتوری',
        '49': 'کاشی و سرامیک',
        '50': 'تجارت عمده وخرده فروشی وسائط نقلیه موتور',
        '51': 'حمل و نقل هوایی',
        '52': 'انبارداری و حمایت از فعالیتهای حمل و نقل',
        '53': 'سیمان، آهک و گچ',
        '54': 'سایر محصولات کانی غیرفلزی',
        '55': 'هتل و رستوران',
        '56': 'سرمایه گذاریها',
        '57': 'بانکها و موسسات اعتباری',
        '58': 'سایر واسطه گریهای مالی',
        '59': 'اوراق حق تقدم استفاده از تسهیلات مسکن',
        '60': 'حمل ونقل، انبارداری و ارتباطات',
        '61': 'حمل و نقل آبی',
        '63': 'فعالیت های پشتیبانی و کمکی حمل و نقل',
        '64': 'مخابرات',
        '65': 'واسطه\u200cگری\u200cهای مالی و پولی',
        '66': 'بیمه وصندوق بازنشستگی به جزتامین اجتماعی',
        '67': 'فعالیتهای کمکی به نهادهای مالی واسط',
        '68': 'صندوق سرمایه گذاری قابل معامله',
        '69': 'اوراق تامین مالی',
        '70': 'انبوه سازی، املاک و مستغلات',
        '71': 'فعالیت مهندسی، تجزیه، تحلیل و آزمایش فنی',
        '72': 'رایانه و فعالیت\u200cهای وابسته به آن',
        '73': 'اطلاعات و ارتباطات',
        '74': 'خدمات فنی و مهندسی',
        '76': 'اوراق بهادار مبتنی بر دارایی فکری',
        '77': 'فعالبت های اجاره و لیزینگ',
        '82': 'فعالیت پشتیبانی اجرائی اداری وحمایت کسب',
        '90': 'فعالیت های هنری، سرگرمی و خلاقانه',
        '93': 'فعالیتهای فرهنگی و ورزشی',
        '98': 'گروه اوراق غیرفعال',
        'X1': 'شاخص',
    }


@file('industrial_groups_overview.html')
async def test_industrial_groups_overview():
    df = await industrial_groups_overview()
    assert [*df.dtypes.items()] == [
        ('group', dtype('O')),
        (':-2', dtype('int64')),
        ('-2:0', dtype('int64')),
        ('0:2', dtype('int64')),
        ('2:', dtype('int64')),
    ]
    assert len(df) > 40


@file('weatherforecast.json')
async def test_market_map_data():
    df = await market_map_data()
    assert len(df) > 300
    assert not df.lVal18AFC.str.contains('ي').any()
    assert [*df.dtypes.items()] == [
        ('insCode', dtype('int64')),
        ('dEven', dtype('int64')),
        ('hEven', dtype('int64')),
        ('pClosing', dtype('float64')),
        ('pDrCotVal', dtype('float64')),
        ('zTotTran', dtype('float64')),
        ('qTotTran5J', dtype('float64')),
        ('qTotCap', dtype('float64')),
        ('priceYesterday', dtype('float64')),
        ('lVal18AFC', dtype('O')),
        ('lVal30', dtype('O')),
        ('lSecVal', dtype('O')),
        ('percent', dtype('float64')),
        ('priceChangePercent', dtype('float64')),
        ('hEvenShow', dtype('O')),
        ('color', dtype('O')),
        ('fontSize', dtype('int64')),
        ('fontColor', dtype('O')),
        ('customLabel', dtype('O')),
    ]


@file('major_holders_activity.html')
async def test_major_holders_activity():
    df = await major_holders_activity()
    dtypes = [*df.dtypes.items()]
    assert dtypes[:3] == [
        ('ins_code', dtype('int64')),
        ('l30', dtype('O')),
        ('holder', dtype('O')),
    ]
    for _, t in dtypes[3:]:
        assert dtype(t) == dtype('float64')


@file('top_industry_groups.html')
async def test_top_industry_groups():
    df = await top_industry_groups()
    assert [*df.dtypes.items()] == [
        ('group', dtype('O')),
        ('mv', dtype('float64')),
        ('tno', dtype('int64')),
        ('tvol', dtype('float64')),
        ('tval', dtype('float64')),
    ]


@file('market_overview.json')
async def test_market_overview():
    d = await market_overview()
    assert_dict_type(d, _MarketOverview)



from logging import info, warning

from tsetmc import _DataFrame
from tsetmc.instruments import Instrument as _Instrument, _LazyDS
from tsetmc.market_watch import market_watch_init as _market_watch_init

# cs == 69: اوراق تامين مالي
# cs == 59: اوراق حق تقدم استفاده از تسهيلات مسكن
_CS_EXCLUSIONS = {59, 69}

# see dev/tsetmc_source_files/market_watch.html
_YVAL_EXCLUSIONS = {
    # OraghMosharekat
    306,
    301,
    706,
    208,
    701,
    # گواهی سسپرده کالایی (سیمان/زعفران)
    327,
}


def _dump(df):
    if not df['code'].is_unique:
        dups = df['code'].duplicated(keep='last')
        warning('%d non-unique code were found', len(dups))
        # Some of the newly added codes are not unique, remove old duplicates.
        df = df[~dups]

    assert df['l18'].is_unique

    df.sort_values('l18', inplace=True)
    df.to_csv(
        _LazyDS.path, index=False, encoding='utf-8-sig', lineterminator='\n'
    )


async def add_instrument(inst: _Instrument) -> None:
    # usually used in conjunction with Instrument.from_search
    code = inst.code
    info = await inst.info()
    df = _LazyDS.df
    l18, l30 = info['lVal18AFC'], info['lVal30']
    if (l1830 := _LazyDS.l18_l130(code)) is not None:
        if l1830 == (l18, l30):
            warning(f'{l18 = } already exists in dataset')
            return
    df.loc[len(df)] = [code, l18, l30]
    _dump(df)


async def update() -> None:
    mwi = await _market_watch_init(market_state=False, best_limits=False)
    prices = mwi['prices']
    # flow == 3: futures market
    prices = prices.query(
        'flow != 3 '
        'and cs not in @_CS_EXCLUSIONS '
        'and yval not in @_YVAL_EXCLUSIONS'
    )
    ds = _LazyDS.df
    merged = _LazyDS._df = ds.merge(
        prices[['l18', 'l30']], how='left', copy=False
    )
    diff = len(merged) - len(ds)
    if diff:
        _dump(merged)
        info(f'{diff} new entries were added by market watch.')
    else:
        info('No new entries were added by market watch.')
